#-------------------------------------------------------
# Asyn PostgresSQL Dispatcher
# Hans Roh (hansroh@gmail.com)
# 2015.6.9
#-------------------------------------------------------
		
DEBUG = False

try:
	import psycopg2
	
except ImportError:	
	class AsynConnect:
		def __init__ (self, address, params = None, lock = None, logger = None):
			logger ("[warn] cannot import psycopg2")
			raise ImportError ("cannot import psycopg2")

else:	
	import asyncore
	from . import dbconnect

	from psycopg2.extensions import POLL_OK, POLL_WRITE, POLL_READ
	_STATE_OK = (POLL_OK, POLL_WRITE, POLL_READ)
		
	class AsynConnect (dbconnect.AsynDBConnect, asyncore.dispatcher):
		def __init__ (self, address, params = None, lock = None, logger = None):
			dbconnect.AsynDBConnect.__init__ (self, address, params, lock, logger)			
			self.conn = None
			self.cur = None			
			asyncore.dispatcher.__init__ (self)
			
		def check_state (self, state):
			if state not in (_STATE_OK):				
				self.logger ("[warn] psycopg2.poll() returned %s" % state)
				self.handle_close ()
		
		def poll (self):
			try:
				return self.socket.poll ()
			except psycopg2.OperationalError:
				pass
			except:				
				self.logger.trace ()
			return -1
			
		def writable (self):			
			return self.out_buffer or not self.connected
			
		def readable (self):
			return self.connected and not self.out_buffer
		
		def add_channel (self, map = None):
			return asyncore.dispatcher.add_channel (self, map)
				
		def del_channel (self, map=None):
			fd = self._fileno
			if map is None:
				map = self._map
			if fd in map:
				del map[fd]
		
		def handle_expt_event (self):
			self.handle_expt ()
			
		def handle_connect_event (self):
			state = self.poll ()			
			if state == POLL_OK:	
				self.handle_connect ()
				self.connected = True
				self.connecting = False		
			else:				
				self.check_state (state)
		
		def handle_write_event (self):		
			if not self.connected:
				self.handle_connect_event ()
			else:	
				self.handle_write ()
		
		def handle_expt (self):
			self.handle_close (psycopg2.OperationalError, "socket panic")
			
		def handle_connect (self):
			self.del_channel ()
			self.conn = self.socket
			self.cur = self.conn.cursor()		
			self.set_socket (self.cur.connection)			
					
		def handle_read (self):
			state = self.poll ()
			if self.cur and state == POLL_OK:
				self.set_event_time ()
				self.has_result = True
				self.close_case_with_end_tran ()
			else:
				self.check_state (state)
				
		def handle_write (self):
			state = self.poll ()
			if self.cur and state == POLL_OK:
				self.set_event_time ()				
				self.cur.execute (self.out_buffer)
				self.out_buffer = ""
			else:				
				self.check_state (state)
				
		#-----------------------------------
		# Overriden
		#-----------------------------------
		def close_case (self):			
			if self.request:
				if self.has_result and self.cur.description:					
					self.request.handle_result (self.cur.description, self.exception_class, self.exception_str, self.fetchall ())					
				else:
					self.request.handle_result (None, self.exception_class, self.exception_str, None)
					self.has_result = False
				self.request = None
			self.set_active (False)
			
		def empty_cursor (self):
			if self.has_result:
				try:
					self.fetchall ()
				except psycopg2.ProgrammingError:
					pass				
		
		def fetchall (self):
			result = self.cur.fetchall ()
			self.has_result = False
			return result
							
		def close (self, deactive = 1):
			if self.cur:
				try:
					self.cur.close ()
				except psycopg2.ProgrammingError:
					pass						
				self.cur = None
			if self.conn:	
				self.conn.close ()
				self.conn = None	
				
			dbconnect.AsynDBConnect.close (self, deactive)
			asyncore.dispatcher.close (self)
			
		def connect (self, force = 0):
			host, port = self.address		
			sock = psycopg2.connect (
				dbname = self.dbname,
				user = self.user,
				password = self.password,
				host = host,
				port = port,
				async = 1
			)
			self.set_socket (sock)
		
		def end_tran (self):
			self.del_channel ()

		def begin_tran (self, request):
			dbconnect.AsynDBConnect.begin_tran (self, request)
			self.out_buffer = request.params [0]
								
		def execute (self, request):			
			self.begin_tran (request)			
			if not self.connected:
				self.connect ()				
			else:
				state = self.poll ()
				if state != POLL_OK:
					self.reconnect ()
				else:
					self.add_channel ()



import sys

PY3 = sys.version_info[0] >= 3
if PY3:
	string_types = str,
	integer_types = int,
	text_type = str
	xrange = range
	
	def reraise(tp, value, tb=None):
		if value.__traceback__ is not tb:
			raise value.with_traceback(tb)
		raise value
      
else:
	import __builtin__
	string_types = __builtin__.basestring,
	text_type = __builtin__.unicode
	integer_types = (int, __builtin__.long)
	xrange = __builtin__.xrange
	
	from py2utils import reraise
	
		



from collections import deque
import random

class Queue:
	def __init__ (self):
		self._init ()
		self._req_id = 0
		self._consumed = 0
	
	def qsize (self):
		return len (self.q)
		
	@property
	def req_id (self):
		return self._req_id
			
	def add (self, req):		
		self._add (req)
		self._req_id += 1
		
	def get (self):
		try:
			self._consumed += 1
			return self._get ()
		except IndexError:
			return None
	
	def _init (self):
		self.q = deque ()
	
	def _add (self, req):
		self.q.append (req)
	
	def _get (self):
		return self.q.popleft ()
		

class RandomQueue (Queue):
	def _init (self):
		self.q = []
		
	def _add (self, req):
		lq = len (self.q)
		if lq == 0:
			self.q.append (req)
		else:
			self.q.insert(random.randrange (lq), req)
	
	def _get (self):		
		return self.q.pop (0)
		
	


from .protocols.http import request as http_request, request_handler as http_request_handler
from .protocols.ws import request_handler as ws_request_handler, request as ws_request
from .protocols.grpc import request as grpc_request
from .dbapi import request as dbo_request
from .protocols.proxy import tunnel_handler
from .protocols.http import localstorage as ls

def make_ws (_method, url, params, auth, headers, meta, proxy, logger):
	req = ws_request.Request (url, params, headers, auth, logger, meta)
	if proxy:
		if _method == 'wss':
			handler_class = tunnel_handler.WSSSLProxyTunnelHandler
		else:
			handler_class = tunnel_handler.WSProxyTunnelHandler
	else:		
		handler_class = ws_request_handler.RequestHandler
	req.handler = handler_class	
	return req

content_types = {
	'xml': "text/xml",
	'json': "application/json",	
	'form': "application/x-www-form-urlencoded",
	'nvp': "text/namevalue"	
}
def make_http (_method, url, params, auth, headers, meta, proxy, logger):
	global content_types
	if not headers:
		headers = {}
	if ls.g:
		headers ['Cookie'] = ls.g.get_cookie_as_string (url)
	if proxy and url.startswith ('https://'):
		handler_class = tunnel_handler.SSLProxyTunnelHandler
	else:			
		handler_class = http_request_handler.RequestHandler

	if _method == "rpc":
		rpcmethod, params = params
		req = http_request.XMLRPCRequest (url, rpcmethod, params, headers, auth, logger, meta)
		
	elif _method == "grpc":
		rpcmethod, params = params
		req = grpc_request.GRPCRequest (url, rpcmethod, params, headers, auth, logger, meta)
	
	else:		
		if _method in ("get", "delete"):						
			req = http_request.HTTPRequest (url, _method.upper (), params, headers, auth, logger, meta)
			
		else:	
			if _method in ("post", "put", "patch"):
				ct = None
				for k, v in headers.items ():
					if k.lower () == "content-type":
						ct = v						
						break
				if not ct: raise TypeError ("Content Type Undefined")
			
			elif _method != "upload":
				if _method [:4] == "post":					
					cta, _method = _method [4:], "post"
				elif _method [:4] == "patch":		
					cta, _method = _method [5:], "patch"
				else:
					cta, _method = _method [3:], "put"							
				ct = content_types.get (cta)
				if not ct: raise TypeError ("Content Type Undefined")
				headers ['Content-Type'] = ct
				
			if _method == "upload":
				req = http_request.HTTPMultipartRequest (url, "POST", params, headers, auth, logger, meta)
			else:
				req = http_request.HTTPRequest (url, _method.upper (), params, headers, auth, logger, meta)	
	
	req.handler = handler_class		
	return req

def make_dbo (_method, server, dbmethod, params, dbname, auth, meta, logger):
	return dbo_request.Request (_method [1:], server, dbname, auth, dbmethod, params, None, meta)
	




def tuple_cb (default, callback):
	if not callback:
		return			
	
	if type (callback) is not tuple:
		return callback (default)
			
	lc = len (callback)
	if lc == 2:
		arg = callback [1]
		if type (arg) is dict:
			args, karg = (), arg
		else:
			args, karg = arg, {}
	elif lc == 3:
		args, karg = callback [1:]
	else:
		args, karg = (), {}
	callback [0] (default, *args, **karg)
	



from ...lib.athreads.fifo import await_ts_fifo
			
class http2_producer_fifo (await_ts_fifo):
	# Asyncore await_fifo replacement
	# For resorting, removeing by http2 priority, cacnel and reset features
	# Also can handle producers has 'ready' method
		
	def remove_from (self, stream_id, lst):
		deleted = 0
		for i in range (len (lst)):
			try:
				producer_stream_id = lst [0].stream_id
			except AttributeError:
				pass
			else:
				if producer_stream_id	== stream_id:
					lst.popleft ()
					deleted += 1
			lst.rotate (1)
		return deleted
							
	def remove (self, stream_id):
		with self._lock:
			if self.l:
				self.remove_from (stream_id, self.l)
			if self.r:
				self.remove_from (stream_id, self.r)
			
	def insert (self, index, item):
		if item is None:
			with self._lock:
				self.has_None = True
			return
		
		with self._lock:	
			if self.has_None:
				return # deny adding	
		
		if hasattr (item, 'ready'):
			with self._lock:
				return self.r.append (item)
		
		# insert by priority
		try:
			w1 = item.weight
			d1 = item.depends_on
		except AttributeError:
			with self._lock:
				return self.insert_into (self.l, index, item)
		
		index = 0
		inserted = False
		with self._lock:
			for each in self.l:
				try:
					w2 = each.weight
					d2 = each.depends_on
				except AttributeError:
					pass
				else:
					if d2 >= d1 and w2 < w1:
						self.insert_into (self.l, index, item)
						inserted = True
						break
				index += 1
				
			if not inserted:
				self.l.append (item)
				
